
#ifndef SED_H
#define SED_H

#define VOID void
#define bzero(x,y) memset((x),0,(y))
#define bcopy(FROM,TO,LEN)  memcpy(TO,FROM,LEN)
#define isblank(c) ((c) == ' ' || (c) == '\t')

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include "rx.h"
#include "getopt.h"
#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <malloc.h>

/* Execute the program 'vec' on the current input line. */
// static struct re_registers regs = {0, 0, 0};

/* Struct vector is used to describe a chunk of a compiled sed program.  
 * There is one vector for the main program, and one for each { } pair,
 * and one for the entire program.  For {} blocks, RETURN_[VI] tells where
 * to continue execution after this VECTOR.
 */

struct abc
{
   int a;
};

struct vector
{
   struct sed_cmd *v;
   int v_length;
   int v_allocated;
   struct vector *return_v;
   int return_i;
};

/* Goto structure is used to hold both GOTO's and labels.  There are two
 * separate lists, one of goto's, called 'jumps', and one of labels, called
 * 'labels'.
 * the V element points to the descriptor for the program-chunk in which the
 * goto was encountered.
 * the v_index element counts which element of the vector actually IS the
 * goto/label.  The first element of the vector is zero.
 * the NAME element is the null-terminated name of the label.
 * next is the next goto/label in the list. 
 */

struct sed_label
{
   struct vector *v;
   int v_index;
   char *name;
   struct sed_label *next;
};

/* ADDR_TYPE is zero for a null address,
 *  one if addr_number is valid, or
 * two if addr_regex is valid,
 * three, if the address is '$'
 * Other values are undefined.
 */

enum addr_types
{
   addr_is_null = 0,
   addr_is_num = 1,
   addr_is_regex = 2,
   addr_is_last = 3,
   addr_is_mod = 4
};

struct addr
{
   int addr_type;
   struct re_pattern_buffer *addr_regex;
   int addr_number;
   int modulo, offset;
};


/* Aflags:  If the low order bit is set, a1 has been
 * matched; apply this command until a2 matches.
 * If the next bit is set, apply this command to all
 * lines that DON'T match the address(es).
 */

#define A1_MATCHED_BIT   01
#define ADDR_BANG_BIT   02

struct sed_cmd
{
   struct addr a1, a2;
   int aflags;
   char cmd;
   union
   {
      /* This structure is used for a, i, and c commands */
      struct { char *text; int text_len; }  cmd_txt;
      /* This is used for b and t commands */
      struct sed_cmd *label;
      /* This for r and w commands */
      FILE *io_file;
      /* This for the hairy s command */
      /* For the flags var:
          low order bit means the 'g' option was given,
          next bit means the 'p' option was given,
          and the next bit means a 'w' option was given,
          and wio_file contains the file to write to. */
#define S_GLOBAL_BIT   01
#define S_PRINT_BIT   02
#define S_WRITE_BIT   04
#define S_NUM_BIT   010
      struct {
         struct re_pattern_buffer *regx;
         char *replacement;
         int replace_length;
         int flags;
         int numb;
         FILE *wio_file;
      } cmd_regex;
      /* This for the y command */
      unsigned char *translate;
      /* For { */
      struct vector *sub;
      /* for t and b */
      struct sed_label *jump;
   } x;
};

/* Sed operates a line at a time. */
struct line
{
   char *text;         /* Pointer to line allocated by malloc. */
   int length;         /* Length of text. */
   int alloc;         /* Allocated space for text. */
};

/* This structure holds information about files opend by the 'r', 'w',
   and 's///w' commands.  In paticular, it holds the FILE pointers to
   use, the file's name. */

#define NUM_FPS   32
struct FilePtrs
{
   FILE *for_read;
   FILE *for_write;
   char *name;
};

void panic (char *str,...);
char *__fp_name (FILE * fp);
FILE *ck_fopen (char *name, char *mode);
void ck_fwrite (char *ptr, int size, int nmemb, FILE * stream);
void ck_fclose (FILE * stream);
VOID *ck_malloc (int size);
VOID *ck_realloc (VOID * ptr, int size);
char *ck_strdup (char *str);
VOID *init_buffer (void);
void flush_buffer (VOID * bb);
int size_buffer (VOID * b);
void add_buffer (VOID * bb, char *p, int n);
void add1_buffer (VOID * bb, int ch);
char *get_buffer (VOID * bb);

void bad_prog (char *why);
void read_file (char *name);
int read_pattern_space_string (void);
void line_copy (struct line * from, struct line * to);
void line_append (struct line * from, struct line * to);
void str_append (struct line * to, char *string, int length);

static struct option longopts[] =
{
   {"expression", 1, NULL, 'e'},
   {"file", 1, NULL, 'f'},
   {"quiet", 0, NULL, 'n'},
   {"silent", 0, NULL, 'n'},
   {"version", 0, NULL, 'V'},
   {"help", 0, NULL, 'h'},
   {NULL, 0, NULL, 0}
};

#endif
